unit UKart;

interface

uses
  ExtCtrls, Graphics, Math, Types, DateUtils, SysUtils, Contnrs, Forms,
  UTrack2D, UState, UPlayer2D, UUtils;

type
  TKart = class(TPlayer2D)
  private
    // Computed params:
    FTurn, FPower, FBattery: integer;
    // Fixed params:
    FUseBattery, FCanRunOut: integer;
    // Free params:
    FPushTurn, FPushPower: integer;
    // MicroState components:
    procedure SetTurn(v: double);
    function  GetTurn: double;
    procedure SetPower(v: double);
    function  GetPower: double;
    procedure SetBattery(v: double);
    function  GetBattery: double;
    // Limits of MicroStates:
    procedure SetMaxPower(v: double);
    function  GetMaxPower: double;
    procedure SetMinPower(v: double);
    function  GetMinPower: double;
    procedure SetMaxTurn(v: double);
    function  GetMaxTurn: double;
    // Fixed Params:
    procedure SetUseBattery(v: boolean);
    function  GetUseBattery: boolean;
    procedure SetCanRunOut(v: boolean);
    function  GetCanRunOut: boolean;
    // Free params:
    function  GetPushTurn: double;
    function  GetPushPower: double;
  protected
    procedure DefineStateParams; override;
    // Update the points on screen that form the player's siluette:
    procedure CalcNodes; override;
    // I dont want accelerae/brake futures to be shown, only left/rigth:
    function  DebugThisFuture: boolean; override;
  public
    OffTrack: boolean;
    constructor create(x, y, a: double);
    procedure Simulate(ms: integer); override;
    procedure draw; override;
    function  DebugColor(ThisFuture: TState): TColor; override;
    // Easy access to State components
    // ===============================
    // Computed params:
    property Turn:    double read GetTurn    write SetTurn;
    property Power:   double read GetPower   write SetPower;
    property Battery: double read GetBattery write SetBattery;
    // Limits of computed params are kind of fixed params:
    property MaxPower:  double read GetMaxPower  write SetMaxPower;
    property MinPower:  double read GetMinPower  write SetMinPower;
    property MaxTurn:   double read GetMaxTurn   write SetMaxTurn;
    // Fixed params:
    property UseBattery: boolean read GetUseBattery write SetUseBattery;
    property CanRunOut:  boolean read GetCanRunOut  write SetCanRunOut;
    // Free params (read only):
    property PushTurn:  double  read GetPushTurn;
    property PushPower: double  read GetPushPower;
  end;


implementation

uses UPlayer, UTrack;


// ****************************
// **  ADMINISTRATIVE STUFF  **
// ****************************

constructor TKart.create(x, y, a: double);
begin
  inherited create;
  Size:= 8;
  PlaceHere(x, y, a);
  // Remember my initial state:
  SetIniState;
  fTargetFutures:= 40;
end;

// Add params to the micro-state definition:
// Format to create a new param: (Name, Value, Max, Min, Precision)
procedure TKart.DefineStateParams;
begin
  // PosX and PosY are mandatory, so they are defined on TPLayer2D:
  inherited;
  // Computed params:
  // ================
  State.AddParam( TStateParam.create('Turn',   ptComputed,   0.0,  PI/4, -PI/4, 0.0) );
  State.AddParam( TStateParam.create('Power',  ptComputed,   0.0,  10.0, -10.0, 0.0) );
  State.AddParam( TStateParam.create('Battery',ptComputed, 100.0, 100.0,   0.0, 0.0) );
  // Indexes:
  FTurn:=    State.ParamIndexByName('Turn');
  FPower:=   State.ParamIndexByName('Power');
  FBattery:= State.ParamIndexByName('Battery');
  // Fixed params:
  // =============
  // Activate use of battery.
  State.AddParam( TStateParam.create('UseBattery', ptFixed, 0, 0, 0, 0) );
  // Can i run out of track? Usually can't, but once you are off, better come back.
  State.AddParam( TStateParam.create('CanRunOut', ptFixed, 0, 0, 0, 0) );
  // Indexes:
  FUseBattery:= State.ParamIndexByName('UseBattery');
  FCanRunOut:=  State.ParamIndexByName('CanRunOut');
  // Free params:
  // ============
  // Max speed you can push the driving wheel left or right (in rad/s)
  // Example: PI/4 = 45/s, so you need 2 s. to change from 45 left to 45 right.
  // Precision of value/2 means 4 possible initial decisions on this free param.
  State.AddParam( TStateParam.create('PushTurn',  ptFree, 0.0,  10,  -10,  5) );
  State.AddParam( TStateParam.create('PushPower', ptFree, 0.0, 180, -180, 50) );
  // Indexes:
  FPushTurn:=  State.ParamIndexByName('PushTurn');
  FPushPower:= State.ParamIndexByName('PushPower');
end;

// *******************
// ** SET/GET STUFF **
// *******************

procedure TKart.SetTurn(v: double);
begin
  State.Param[FTurn].Value:= v;
end;
function  TKart.GetTurn: double;
begin
  result:= State.Param[FTurn].Value;
end;

procedure TKart.SetPower(v: double);
begin
  State.Param[FPower].Value:= v;
end;
function  TKart.GetPower: double;
begin
  result:= State.Param[FPower].Value;
end;

procedure TKart.SetBattery(v: double);
begin
  State.Param[FBattery].Value:= v;
end;
function  TKart.GetBattery: double;
begin
  result:= State.Param[FBattery].Value;
end;

procedure TKart.SetMaxPower(v: double);
begin
  State.Param[FPower].MaxValue:= v;
end;
function  TKart.GetMaxPower: double;
begin
  result:= State.Param[FPower].MaxValue;
end;

procedure TKart.SetMinPower(v: double);
begin
  State.Param[FPower].MinValue:= v;
end;
function  TKart.GetMinPower: double;
begin
  result:= State.Param[FPower].MinValue;
end;

procedure TKart.SetMaxTurn(v: double);
begin
  State.Param[FTurn].MaxValue:= v;
end;
function  TKart.GetMaxTurn: double;
begin
  result:= State.Param[FTurn].MaxValue;
end;

procedure TKart.SetUseBattery(v: boolean);
begin
  if v then
    State.Param[FUseBattery].Value:= 1
  else
    State.Param[FUseBattery].Value:= 0;
end;
function  TKart.GetUseBattery: boolean;
begin
  result:= (State.Param[FUseBattery].Value>0.5);
end;

procedure TKart.SetCanRunOut(v: boolean);
begin
  if v then
    State.Param[FCanRunOut].Value:= 1
  else
    State.Param[FCanRunOut].Value:= 0;
end;
function  TKart.GetCanRunOut: boolean;
begin
  result:= (State.Param[FCanRunOut].Value>0.5);
end;

function  TKart.GetPushTurn: double;
begin
  result:= State.Param[FPushTurn].Value;
end;

function  TKart.GetPushPower: double;
begin
  result:= State.Param[FPushPower].Value;
end;


// *********************
// **  DRAWING STUFF  **
// *********************

procedure TKart.CalcNodes;
var
  i: integer;
  X, Y: double;
const
  Angles: array[0..3] of double = (-0.1, 0.1, 0.8, -0.8);
begin
  // Nodes is a dynamic array, so I must check its size and if it is
  if NodesNeedIni then begin
    SetLength(Nodes, 4);
    NodesNeedIni:= false;
    NodesNeedRecalc:= true;
  end;
  // If nothing changed from last time, don't waste time:
  if NodesNeedRecalc then begin
    for i:= Low(Nodes) to High(Nodes) do begin
      X:= PosX + Size*cos(Angle + Angles[i]*PI);
      Nodes[i].X:= round(X);
      Y:= PosY + Size*sin(Angle + Angles[i]*PI);
      Nodes[i].Y:= round(Y);
    end;
  end;
  NodesNeedRecalc:= false;
end;

procedure TKart.draw;
var
  i: integer;
  Diam: integer;
  mx, my, Joy_X, Joy_Y: double;
  TyreIni, TyreEnd: TPoint;
  Track2D: TTrack2D;
begin
  // TPlayer will draw the body for me using the nodes:
  inherited;
  Track2D:= TTrack2D(Track);
  // Add the 2 front tires only (so you can see where is it heading to):
  for i:= 0 to 1 do begin
    // Tyres are just lines:
    TyreIni.X:= round(Self.PosX + Size*(cos(Angle + (i*2-1) * 0.2*PI)));
    TyreIni.Y:= round(Self.PosY + Size*(sin(Angle + (i*2-1) * 0.2*PI)));
    TyreEnd.X:= TyreIni.X + round(Size*0.6*(cos(Angle + Turn)));
    TyreEnd.Y:= TyreIni.Y + round(Size*0.6*(sin(Angle + Turn)));
    // Draw it:
    Track2D.Img.Canvas.Pen.Width:= 2;
    Track2D.Img.Canvas.MoveTo(TyreIni.X, TyreIni.Y);
    Track2D.Img.Canvas.LineTo(TyreEnd.X, TyreEnd.Y);
    Track2D.Img.Canvas.Pen.Width:= 1;
  end;
  // Draw joystick position:
  Track2D.Img.Canvas.Brush.Color:= clWhite;
  // The position change with power and turn:
  if (Power>=0) then
    mx:= Power/MaxPower
  else
    mx:= Power/Abs(MinPower);
  my:= Turn/MaxTurn;
  Joy_X:= PosX + (cos(Angle)*mx*0.6 - sin(Angle)*my) * Size;
  Joy_Y:= PosY + (sin(Angle)*mx*0.6 + cos(Angle)*my) * Size;
  Diam:= round(Size * 0.4);
  // Draw it:
  // Stick...
  Track2D.Img.Canvas.Pen.Width:= 2;
  Track2D.Img.Canvas.MoveTo(round(PosX), round(PosY));
  Track2D.Img.Canvas.LineTo(round(Joy_X), round(Joy_Y));
  Track2D.Img.Canvas.Pen.Width:= 1;
  // Ball...
  // Fill it red (power<0) or green (power>0):
  if (Power>=0) then begin
    //Green to black
    i:= round(255 * Sqrt(Power/MaxPower));
    Track2D.Img.Canvas.Brush.Color:= RGBToColor(255-i, 255, 255-i);
  end else begin
    //Yellow to black
    i:= round(255 * Sqrt(Power/MinPower));
    Track2D.Img.Canvas.Brush.Color:= RGBToColor(255, 255, 255-i);
  end;
  Track2D.Img.Canvas.Pen.Color:=   clLtGray;
  Track2D.Img.Canvas.Ellipse(round(Joy_X-Diam), round(Joy_Y-Diam), round(Joy_X+Diam), round(Joy_Y+Diam));
  Track2D.Img.Canvas.Pen.Color:=   clWhite;
  Track2D.Img.Canvas.Brush.Color:= clWhite;
  Track2D.Img.Canvas.Ellipse(round(Joy_X-Diam*0.8), round(Joy_Y-Diam*0.8),
                           round(Joy_X+Diam*0.2), round(Joy_Y+Diam*0.2));
  Track2D.Img.Canvas.Brush.Style:= bsClear;
  Track2D.Img.Canvas.Pen.Color:=   clBlack;
  Track2D.Img.Canvas.Ellipse(round(Joy_X-Diam), round(Joy_Y-Diam), round(Joy_X+Diam), round(Joy_Y+Diam));
  // Add a Battery level:
  if UseBattery then begin
    Track2D.Img.Canvas.Pen.Color:= clBlack;
    Track2D.Img.Canvas.Brush.Color:= clLtGray;
    Track2D.Img.Canvas.Rectangle(round(PosX+Size), round(PosY+Size), round(PosX+Size+3), round(PosY-Size));
    if Battery < 25 then
      Track2D.Img.Canvas.Pen.Color:= clRed
    else if Battery < 50 then
      Track2D.Img.Canvas.Pen.Color:= clYellow
    else
      Track2D.Img.Canvas.Pen.Color:= clLime;
    Track2D.Img.Canvas.MoveTo(round(PosX+Size+1), round(PosY+Size-1));
    Track2D.Img.Canvas.LineTo(round(PosX+Size+1), round(PosY+Size-1-(2*Size-2)*Battery/100));
  end;
end;

// Debug only left/right decision, not run/stop ones:
function TKart.DebugThisFuture: boolean;
begin
  result:= (State.Param[FPushTurn].IniValue <> 0);
end;

function TKart.DebugColor(ThisFuture: TState): TColor;
var C1, C2: integer;
begin
  // Red or blue base color proportionally to left/right push:
  //C1:= Round( 255 * Max(0, Abs(ThisFuture.Param[FPushTurn].IniValue) / ThisFuture.Param[FPushTurn].MaxValue));
  // Better use plain red or blue:
  C1:= 255;
  // As future goes away it fades to white:
  C2:= Round( 55 + 200 * ThisFuture.TimePassed / TTrack(Track).FutureMSeconds);
  if (ThisFuture.Param[FPushTurn].IniValue > 0) then
    // Right turning is red:
    result:= RGBToColor(C1, C2, C2)
  else if (ThisFuture.Param[FPushTurn].IniValue < 0) then
    // Left turning is blue:
    result:= RGBToColor(C2, C2, C1)
  else
    // Center options are grey:
    result:= RGBToColor(C2, C2, C2);
end;


// ***********************
// **  KART SIMULATION  **
// ***********************

// Let run the simulation for a 1/10 of the game tick time (1/FPS, btw)
procedure TKart.Simulate(ms: integer);
var
  // Time pased
  s, s2: double;
  // Acceleration components:
  Acc_X, Acc_Y: double;
  // Need vectors to decompose velocity:
  Va_L, Va_X, Va_Y, Vn_L, Vn_X, Vn_Y: double;
  // Friction coefs. for both velocities:
  UseFriction, UseSideFriction: double;
  Friction: double;
begin
  // Inherited will store old values for me:
  inherited;
  // OFF-TRACK CONTROL
  // =================
  // Are you out of track now?
  OffTrack:= not IsInsideTrack;
  // If it is a real move, set the posibility to run offtrack in the test futures:
  if not Imagining then
    CanRunOut:= OffTrack
  else if not OffTrack then
    // A future is back to track, don't allow it to leave again easily!
    CanRunOut:= false;
  // Leaving the screen is really bad, avoid it:
  if not IsInsideImage then begin
    State.MyScore:= 0;
  end;
  // In the moment you leave track, the future "truncates" and simulation stops.:
  if OffTrack and not CanRunOut then begin
    // Futures truncate, real moves don't:
    if Imagining then begin
      dead:= true; // Do not go on simulating this future.
      // On AI level 1, crashes count as 0!
      if AI_Level = 1 then
        State.MyScore:= 0;
      exit;
    end else begin
      //  Here you could take some health form the player
      Power:= 0;
    end;
  end;
  // TIMES
  // =====
  // Milliseconds to seconds:
  s:= (ms/1000);
  s2:= Sqr(s)/2;
  // APPLY DECISION
  // ==============
  // Dead karts -after a crash or going out track- dont push power!
  if dead then begin
    Power:= 0;
  end else begin
    Turn:=  Turn  + PushTurn  * s;
    Power:= Power + PushPower * s;
  end;
  // VELOCITY
  // ========
  // Decompose vel. into "ahead vel" (Va) and normal vel (Vn):
  // Ahead vel.: Length is first calc, then get its components:
  Va_L:= VelX * cos(Angle) + VelY * sin(Angle);
  Va_X:= Va_L * cos(Angle);
  Va_Y:= Va_L * sin(Angle);
  // Now the "rest" of vel. will be "normal velocity":
  Vn_X:= VelX - Va_X;
  Vn_Y:= VelY - Va_Y;
  // FRICTION COEFS
  // ==============
  UseFriction:=     TTrack2D(Track).Friction;
  UseSideFriction:= TTrack2D(Track).SideFriction;
  // Karts out of track suffer a higer friction
  if OffTrack then begin
    UseFriction:=     UseFriction * 5;
    UseSideFriction:= UseSideFriction * 3;
  end;
  // Convert to a 0-1 coeficient:
  Friction:=  (1 + 0.01*Va_L) *  UseFriction;
  Friction:=  Max(0, (1- (ms/100 * Friction )));
  // BRAKES
  // ======
  // Braking takes out a fixed amount to the velocity length:
  if (Power<0) then begin
    // At full braking, I take off 100 times the friction on the velocity.
    Va_L:= Max(0, Va_L - s*UseFriction* 100 *MaxPower*(Abs(Power/MinPower)));
    Va_X:= Va_L * cos(Angle);
    Va_Y:= Va_L * sin(Angle);
  end;
  // SLIDING CONTROL
  // ===============
  // Is like braking but with a fixed "power" and a maximum g.
  Vn_L:= Sqrt(Sqr(Vn_X) + Sqr(Vn_Y));
  if (Vn_L>0.0001) then begin
    // Normalize:
    Vn_X:= Vn_X / Vn_L;
    Vn_Y:= Vn_Y / Vn_L;
    // Adjust length:
    Vn_L:= Max(0, Vn_L - 100*s*UseSideFriction);
    // Recalc componets:
    Vn_X:= Vn_X * Vn_L;
    Vn_Y:= Vn_Y * Vn_L;
  end;
  // UPDATES
  // =======
  // Angle:
  Angle:= Angle + Turn * s * Min(1, Va_L/10);
  // Acceleration:
  Acc_X:= Max(0, Power) * cos(Angle + Turn/2);
  Acc_Y:= Max(0, Power) * sin(Angle + Turn/2);
  // Velocity:
  VelX:= (Va_X*Friction + Vn_X) + s * Acc_X;
  VelY:= (Va_Y*Friction + Vn_Y) + s * Acc_Y;
  // Position:
  // Remember: Pos(t+s) = Pos(t) + Vel(t)*s + Acc(t)*s^2
  PosX:= PosX + VelX * s + Acc_X * s2;
  PosY:= PosY + VelY * s + Acc_Y * s2;

  {
  // SCORING MOVEMENT
  // ================
  // On AI levels 1 to 3 we don't use raced distance
  if (AI_Level <= 3) then begin
    State.MyScore:= 1;
  end else begin
    // Add the distance raced in this move:
    NewRaced:= Sqrt(Sqr(PosX-OldPosX)+Sqr(PosY-OldPosY)) / Size;
    // While offtrack, distance score much lower, so going back to track is forced:
    if OffTrack then
      NewRaced:= NewRaced/100;
    // Add scoring of this delta move into the state's scoring:
    State.AddMyScore(NewRaced);
  end;
  }

  // Update 4 nodes of the kart skeleton:
  NodesNeedRecalc:= true;
end;

end.
