unit UPlayer;

interface

uses Graphics, Types, Contnrs,
     UState, UListOfFutures, UGoal;

// Higher level of AI possible:
const Max_AI_Level : integer = 6;

type
  TPlayer = class(TObject)
  private
    // Team the player belongs to:
    fTeam: TObject;
    // Color for drawing:
    fColor: TColor;
    // Debug flags:
    fDebugTraces: boolean;
    fDebugOptions: boolean;
    fDebugGoals: boolean;
    // Debuging options' score is tricky, some times range from 0 to 1000.
    fFutureScoreScale: double;
    // Am I imagining futures by taking random decisions?
    // When passing from false to true you must "initialize" FutureState, and
    // when setting it back to false, you must store the future found in your
    // internal list.
    fImagining: boolean;
    // Witch things motivate this player to move here or there?
    fGoals: TObjectList;
    // For testing purpouses you can downgrade AI to previous versions:
    // Level 1 : Future.FinalScore = 1 (except crashing futures that score 0)
    // Level 2 : Future.FinalScore = 1 always.
    // Level 3 : Future.FinalScore = R (raced distance)
    // Level 4 : Future.FinalScore = R^2 (best formulae found so far)
    // Level 5 : Splines are used to help on Y-shaped situations (RefinedAI).
    fAI_Level: integer;
    //
    // ************************
    // **  SETTER & GETTERS  **
    // ************************
    //
    // Actual state of the player (real or imaginary for futures):
    function GetState: TState;
    function GetTrack: TObject;
    function GetGoalCount: integer;
    //
    // MicroState components:
    // ======================
    procedure SetDead(v: boolean);
    function  GetDead: boolean;
    procedure SetSize(v: double);
    function  GetSize: double;
    procedure SetImagining(v: boolean);
    function  GetGridSize: integer;
    procedure SetGridSize(v: integer);
    function  GetUseRefinedAI: boolean;
  protected
    // A "State" contains all parameters (time-changing, fixed, AI driven) the player.
    // There are 2 ones: the actual state of the "real" player, and the "imaginary"
    // state used to imagine a future for this player. I will use "State" in the
    // code, but it will mask one of the two possible states depending on begin
    // really deciding with futures previously found, or you are precisely looking
    // for those futures.
    fActualState, fFutureState: TState;
    // Index of the StateParams that are common to all TPlayers:
    fSize, fDead, fLoad, fHealth: integer;
    // Grid size to convert from micro to macro state.
    // Computed params with a precision will change it as you redefine grid size.
    // If you define a target number of futures needed to decide properly, then
    // "AutoGridSize" will make grid size to change so the number of futures found
    // tend to fit the desired number (thinks better).
    fGridSize, fMaxGridSize, fMinGridSize: integer;
    fAutoGridSize: boolean;
    fTargetFutures: integer;
    NodesNeedIni: boolean;
    NodesNeedRecalc: boolean;
    function  DebugThisFuture: boolean; virtual;
    procedure CalcNodes; virtual; abstract;
    function  GetGoal(m: integer): TGoal;
    property FutureScoreScale: double read fFutureScoreScale write fFutureScoreScale;
    procedure DefineStateParams; virtual;
  public
    // List of futures found:
    Futures: TListOfFutures;
    //
    constructor create;
    destructor  destroy; override;
    // Back to an initial state:
    procedure ResetState; virtual; //State  = Position, velocity... (change with time)
    // Add a new goal to the player's AI:
    procedure AddGoal(v: TGoal);
    // Copy params from another TPlayer:
    procedure CopyFrom(Other: TPlayer); virtual;
    // Drawing stuff:
    procedure Draw; virtual;
    function  DebugColor(ThisFuture: TState): TColor; virtual;
    procedure DrawFuturePoint; virtual; abstract; // Level 1 debug
    procedure DrawFutureScore(f: TState); virtual; abstract; // Level 3 debug
    procedure DrawGoals; virtual;
    // Grid size control:
    property GridSize: integer read GetGridSize write SetGridSize;
    property AutoGridSize: boolean read fAutoGridSize write fAutoGridSize;
    // After taking a decision, adjust grid size so number of futures found is ok:
    procedure AutoAdjustGridSize; virtual;
    //
    // Save actual state as the intial one:
    procedure SetIniState; virtual;
    property Team:  TObject read fTeam write fTeam;
    property Track: TObject read GetTrack;
    // Debug different things:
    property DebugTraces:  boolean read fDebugTraces  write fDebugTraces;
    property DebugOptions: boolean read fDebugOptions write fDebugOptions;
    property DebugGoals:   boolean read fDebugGoals   write fDebugGoals;
    // Color of the player:
    property Color: TColor read FColor write FColor;
    // SIMULATION:
    // ===========
    procedure Simulate(ms: integer); virtual;
    procedure ScoreGoals;
    function  CheckCollision(Other: TPlayer): boolean; virtual; abstract;
    // Witch is my state now? (the real or the imaginary one):
    property  State: TState read GetState;
    // Delete all futures as I am going to imagine a new set from my actual state:
    procedure ResetFutures;
    // Start/Stop imagining a new future involves managing the list of futures:
    property  Imagining: boolean read FImagining write SetImagining;
    // Taking decisions before moving:
    // Take the first random decision of a new future:
    procedure TakeFirstDecision;
    // Take another random decision to continue with a future:
    procedure TakeRandomDecision;
    // Take an intelligent decision based on futures found and their scorings:
    procedure TakeIntelligentDecision;
    // Get a goal by index number:
    property Goal[m: integer]: TGoal read GetGoal;
    property GoalCount: integer read GetGoalCount;
    // Debug it:
    procedure DrawOptions;
    // AI level (1 to 4 by now).
    property AI_Level: integer read fAI_Level write fAI_Level;
    // Use refined decision layer to avoid Y shaped situations (AILevel=4).
    property UseRefinedAI: boolean read GetUseRefinedAI;
    property Size: double read GetSize write SetSize;
    property Dead: boolean read GetDead write SetDead;
  end;

implementation

uses Math, UUtils, UTrack, UTeam;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TPlayer.create;
begin
  // Create the underlying object:
  inherited;
  // State:
  fActualState:= TState.create(self);
  fFutureState:= TState.create(self);
  // Fill with all the components I need (position, angle, etc):
  DefineStateParams();
  // Futures:
  Futures:= TListOfFutures.create(self);
  // AI Level:
  AI_Level:= Max_AI_Level;
  // Personal goals in life:
  fGoals:= TObjectList.create;
  fGoals.OwnsObjects:= true;
  // Grid size control:
  fGridSize:=    10;
  fMinGridSize:=  5;
  fMaxGridSize:= 30;
  fAutoGridSize:= true;
  fTargetFutures:= 50;
  // Set internal params:
  NodesNeedRecalc:= true;
  NodesNeedIni:= true;
  Imagining:= false;
  DebugTraces:= false;
  DebugOptions:= false;
  DebugGoals:= false;
  FutureScoreScale:= 1;
  // Set a random color:
  Color:= RGBToColor(100+round(random*155), 100+round(random*155), 100+round(random*155));
  SetIniState;
end;

destructor TPlayer.destroy;
begin
  Futures.Free;
  fActualState.Free;
  fFutureState.Free;
  fGoals.Free;
  inherited;
end;

procedure TPlayer.CopyFrom(Other: TPlayer);
begin
  // Copy its State:
  fActualState.CopyFrom(Other.fActualState);
  // Reset initial state and scoring:
  SetIniState;
  // Thinking params:
  Self.Imagining:=    Other.Imagining; // Should be false.
  Self.AI_Level:=     Other.AI_Level;
  Self.AutoGridSize:= Other.AutoGridSize;
end;


// ******************
// ** STATES STUFF **
// ******************

procedure TPlayer.DefineStateParams;
begin
  // I will add items to fActualState, not fFutureState, so I set "Imagining" to
  // false (it should already be so) so State will point to fActualState when used
  // from a descendant's DefineStateParams.
  Imagining:= false;
  // Fixed params: (simulation don't change them):
  State.AddParam( TStateParam.create('Size',   ptFixed,   6.0,   3.0, 25.0, 0.0) ); // Player's size
  State.AddParam( TStateParam.create('Dead',   ptFixed,     0,     0,    1, 0.0) );
  State.AddParam( TStateParam.create('Health', ptFixed, 100.0, 100.0,  0.0, 0.0) );
  State.AddParam( TStateParam.create('Load',   ptFixed,   0.0, 100.0,  0.0, 0.0) ); // Internal storage
  // Index of all created params (for fast access as property):
  fSize:=   State.ParamIndexByName('Size');
  fDead:=   State.ParamIndexByName('Dead');
  fHealth:= State.ParamIndexByName('Health');
  fLoad:=   State.ParamIndexByName('Load');
end;

// Get actual (real or imaginary) state depending on the proccess you are now:
// Moving or imagining new futures.
function TPlayer.GetState: TState;
begin
  if Imagining then
    result:= fFutureState
  else
    result:= fActualState;
end;

function TPlayer.GetTrack: TObject;
begin
  result:= TTeam(fTeam).Track;
end;

// Reset state to initial:
procedure TPlayer.ResetState;
begin
  NodesNeedRecalc:= true;
  fActualState.BackToIniState;
end;

// Set the actual state as the "initial":
procedure TPlayer.SetIniState;
begin
  // Reset scoring:
  State.ResetScore;
  // Copy actual values into ini and old parts of it.
  State.SetIniState();
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TPlayer.Draw;
begin
  // Before drawing you need to recalc node positions:
  CalcNodes;
  // Each kind of TPlayer will have to do the rest.
  // ...
end;

// Draw ending points of the futures:
procedure TPlayer.DrawOptions;
begin
  Futures.DrawOptions;
end;

// Draw the "Goals" that pushes the player:
procedure TPlayer.DrawGoals;
var i: integer;
begin
  if DebugGoals then begin
    for i:= 0 to GoalCount-1 do begin
      if Goal[i].Drawable then begin
        Goal[i].Draw(self);
      end;
    end;
  end;
end;

// Witch color do I use for debbugins this future? Red = left, etc.
function TPlayer.DebugColor(ThisFuture: TState): TColor;
begin
  result:= clGray;
end;

// May be you only want some futures to be shown while debugging:
function TPlayer.DebugThisFuture: boolean;
begin
  result:= true;
end;


// **********************
// ** SIMULATION STUFF **
// **********************

procedure TPlayer.Simulate(ms: integer);
begin
  // Before simulating, store old values:
  State.SetOldState;
  // Time passes by!
  State.TimePassed:= State.TimePassed + ms;
  // Each kind of TPlayer will have to do the rest...
  // ...
end;


// *****************
// ** GOALS STUFF **
// *****************

// After a simulation, the State is changed, and I need to score this change.
procedure TPlayer.ScoreGoals;
var m: integer;
begin
  for m:= 0 to GoalCount-1 do begin
    Goal[m].ScoreGoal(self);
    Goal[m].AddScore(self);
  end;
end;

// Add a new goal to the AI:
procedure TPlayer.AddGoal(v: TGoal);
begin
  fGoals.Add(v);
end;

function TPlayer.GetGoal(m: integer): TGoal;
begin
  result:= TGoal(fGoals[m]);
end;

function TPlayer.GetGoalCount: integer;
begin
  result:= fGoals.count;
end;


// ********************
// ** THINKING STUFF **
// ********************

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeFirstDecision;
begin
  // Dead things (broken karts) don't decide:
  if Dead then exit;
  State.TakeFirstDecision;
end;

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeRandomDecision;
begin
  // Dead things (broken karts) don't decide:
  if Dead then exit;
  State.TakeRandomDecision;
end;

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeIntelligentDecision;
begin
  // Analize futures found and take an intelligent decision based on them.
  Futures.TakeIntelligentDecision(State);
  // Auto adjust grid size to tend to a given number of diff. futures.:
  AutoAdjustGridSize;
end;

// Star/Stop imagining a new future triggers the creation and storing of it
procedure TPlayer.SetImagining(v: boolean);
begin
  // Detect only real changes:
  if (v <> fImagining) then begin
    // Set the new "mind state":
    fImagining:= v;
    // Nodes needs recalc as position has changed:
    NodesNeedRecalc:= true;
    // Starting to imagine a new future or just coming back to reality?
    if Imagining then begin
      // Start of the imagination proccess of a new future:
      //
      // Basically you only need to set your "imaginary state" as the on/used
      // state as this will be the starting point for the future to come:
      fFutureState.CopyFrom(fActualState);
      // Scoring of the future is still empty:
      fFutureState.ResetScore;
    end else begin
      // End of the imagination proccess of a new future:
      //
      // The future is found, store it on my list of futures so I can later
      // decide witch was the best decision and take it on the actual state.
      Futures.AddFuture(fFutureState);
    end;
  end;
end;

// Delete all futures as I am going to imagine a new set from my actual state:
procedure TPlayer.ResetFutures;
begin
  Futures.Clear;
end;

// Use splines to avoid Y-shaped situations? (AI_Level 6)
function TPlayer.GetUseRefinedAI: boolean;
begin
  result:= (AI_Level>=6);
end;


// ***********************
// ** GRID SIZE CONTROL **
// ***********************

// Grid size control:
function  TPlayer.GetGridSize: integer;
begin
  result:= Min(FMaxGridSize, Max(FMinGridSize, FGridSize));
end;

// Modify the precision of computed params so converting from micro to macro
// state generate more or less different futures. If your player needs a more
// specific GridSize to Precision schema, override it, call inherited, then
// redefine the precisions as needed.
procedure TPlayer.SetGridSize(v: integer);
var i:integer;
begin
  fGridSize:= Min(fMaxGridSize, Max(fMinGridSize, v));
  // By default, all computed params with a precision (those that convert from
  // micro to macro state when a future is reached) copy this "GridSize" as its
  // "Precision", so if only PosX and PosY are defined, a grid size is exactly
  // what you should spect in a vectorial graphic editor.
  for i:= 0 to State.MacroParamCount-1 do
    State.MacroParam[i].Precision:= v;
end;

// After taking a decision, adjust grid size so the number of different futures
// tend to a "magical" number. This way, if your are about to crash and you get
// few futures, grid size decreases and futures are more "fine grained" so get
// get more of them and the decision is more stable:
procedure TPlayer.AutoAdjustGridSize;
begin
  // If I have to tend to get 100 futures, and found 50, then decrease grid size:
  // Do it by averaging last grid size with estimated one, so it changes smoothly.
  if AutoGridSize then
    GridSize:= round(GridSize * (1 + Futures.DifferentFutures / fTargetFutures)/2);
end;


// ******************************
// ** SET/GET FOR STATE PARAMS **
// ******************************

// Player's size:
procedure TPlayer.SetSize(v: double);
begin
  if (State.Param[fSize].Value <> v) then begin
    State.Param[fSize].Value:= v;
    NodesNeedRecalc:= true;
  end;
end;
function  TPlayer.GetSize: double;
begin
  result:= State.Param[fSize].Value;
end;

// Dead = broken, no more decisions, no more simulating, your future ends here.
procedure TPlayer.SetDead(v: boolean);
begin
  if v then
    State.Param[fDead].Value:= 1
  else
    State.Param[fDead].Value:= 0;
end;
function  TPlayer.GetDead: boolean;
begin
  result:= (State.Param[fDead].Value>0.5);
end;


end.
