unit UPlayer2D;

interface

uses Types,
     UState, UPlayer;

type
  TPlayer2D = class(TPlayer)
  private
    // State components
    // ================
    procedure SetPosX(v: double);
    function  GetPosX: double;
    procedure SetPosY(v: double);
    function  GetPosY: double;
    procedure SetAngle(v: double);
    function  GetAngle: double;
    function  GetOldPosX: double;
    function  GetOldPosY: double;
    procedure SetVelX(v: double);
    function  GetVelX: double;
    procedure SetVelY(v: double);
    function  GetVelY: double;
  protected
    // Index of the StateParams that are common to all TPlayers:
    FPosX, FPosY, FAngle, FVelX, FVelY: integer;
    function  GetMaxX: double;
    function  GetMinX: double;
    function  GetMaxY: double;
    function  GetMinY: double;
    procedure DefineStateParams; override;
    procedure CalcNodes; override;
  public
    // Those 2D node list forms the body polygon:
    Nodes: array of TPoint;
    // Administrative:
    constructor create;
    // Change ini and actual 2D position:
    procedure PlaceHere(x, y, a: double);
    // Move to a new relative 2D position:
    procedure HandMove(dx, dy, da: double; repaint: boolean = true);
    // Check collisions:
    function  IsInsideImage: boolean;
    function  IsInsideTrack: boolean;
    function  CheckCollision(Other: TPlayer): boolean; override;
    // Draw the player on the track:
    procedure Draw; override;
    procedure DrawBoxLimits;
    procedure DrawBigPoint;
    // Draw debug information:
    procedure DrawFuturePoint; override;
    procedure DrawFutureScore(f: TState); override;

    // 2D box containg the player:
    property MaxX: double read GetMaxX;
    property MinX: double read GetMinX;
    property MaxY: double read GetMaxY;
    property MinY: double read GetMinY;
    // Easy access to State components
    // ===============================
    // State params:
    property PosX:    double read GetPosX  write SetPosX;
    property PosY:    double read GetPosY  write SetPosY;
    property Angle:   double read GetAngle write SetAngle;
    property VelX:    double read GetVelX  write SetVelX;
    property VelY:    double read GetVelY  write SetVelY;
    // Previous state params:
    property OldPosX: double read GetOldPosX;
    property OldPosY: double read GetOldPosY;
  end;

implementation


uses Graphics, Math, Forms,
     UTrack2D, UGoal2DRaced;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TPlayer2D.create;
begin
  inherited;
  // Set an intial position and angle:
  PosX:= 10;
  PosY:= 10;
  Angle:= 0;
  NodesNeedRecalc:= true;
  // Players in 2D has a natural "goal" or "tendence" for racing:
  AddGoal(TGoal2DRaced.create(1));
end;

procedure TPlayer2D.DefineStateParams;
begin
  inherited;
  // A 2D TPlayer must have a PosX and PosY at least:
  // Computed params: (simulation will change them):
  State.AddParam( TStateParam.create('PosX',  ptComputed, 0.0, 0.0, 0.0, 15.0) );
  State.AddParam( TStateParam.create('PosY',  ptComputed, 0.0, 0.0, 0.0, 15.0) );
  // Angle is computed but not used in macro states, so precision is 0.
  State.AddParam( TStateParam.create('Angle', ptComputed, 0.0, 0.0, 0.0, 0.0) );
  State.AddParam( TStateParam.create('VelX',  ptComputed, 0.0, 0.0, 0.0, 0.0) );
  State.AddParam( TStateParam.create('VelY',  ptComputed, 0.0, 0.0, 0.0, 0.0) );
  // Index of all created params (for fast access as property):
  FPosX:=  State.ParamIndexByName('PosX');
  FPosY:=  State.ParamIndexByName('PosY');
  FAngle:= State.ParamIndexByName('Angle');
  FVelX:=  State.ParamIndexByName('VelX');
  FVelY:=  State.ParamIndexByName('VelY');
end;


// **********************
// ** SIMULATION STUFF **
// **********************

function TPlayer2D.IsInsideImage: boolean;
var i: integer;
begin
  CalcNodes;
  // Ask the Track if the player's node positions are inside image.
  for i:= Low(Nodes) to High(Nodes) do begin
    if not TTrack2D(Track).IsInsideImage(Nodes[i].X, Nodes[i].Y) then begin
      result:= false;
      exit;
    end;
  end;
  result:= true;
end;

function TPlayer2D.IsInsideTrack: boolean;
var i: integer;
begin
  CalcNodes;
  // Ask the Track if the player's position is on track or outside it.
  // Instead of center point, use all of the 4 points of the kart skeleton.
  for i:= Low(Nodes) to High(Nodes) do begin
    // Two ways: One point is out then player is out, or if one point is in then
    // the player is in. Both are ok, but I prefer allowing players to be on the
    // adge of the track.
    //
    // Mode 1: One point is off track, then player is off track
    if not TTrack2d(Track).IsInsideTrack(Nodes[i].X, Nodes[i].Y) then begin
      result:= false;
      exit;
    end;
    // Mode 2: One point is on track, then player is on track
    //if TTrack(Track).IsInsideTrack(Nodes[i].X, Nodes[i].Y) then begin
    //  result:= true;
    //  exit;
    //end;
  end;
  result:= true;
end;

// 2 Players collide when a node of one is inside of the other's poligon.
// Point inside a convex poligon? All cross-products must be same signed.
function TPlayer2D.CheckCollision(Other: TPlayer): boolean;
var
  Other2D: TPlayer2D;
  OtherVector: array of TPoint;
  CrossProduct: double;
  i, j, nexti, s: integer;
  ItIsInside: boolean;
  Side: integer;
begin
  result:= false;
  // Only Player2D can be checked:
  if (Other is TPlayer2D) then
    Other2D:= TPlayer2D(Other)
  else
    exit;
  // Force node recalculation if necessary:
  Self.CalcNodes;
  Other2D.CalcNodes;

  {
  if (debug=3) then begin
    TTrack2D(Track).ClearTrack;
    TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
    // Draw my lines:
    TTrack2D(Track).Img.Canvas.Pen.Width:= 2;
    TTrack2D(Track).Img.Canvas.Pen.Color:= Self.Color;
    TTrack2D(Track).Img.Canvas.Polygon(Nodes);
    TTrack2D(Track).Img.Canvas.Pen.Width:= 1;
    // Draw others lines:
    TTrack2D(Track).Img.Canvas.Pen.Color:= Other2D.Color;
    TTrack2D(Track).Img.Canvas.Polygon(Other2D.Nodes);
    // Left it prepared to draw points:
    TTrack2D(Track).Img.Canvas.Pen.Color:= clBlack;
    TTrack2D(Track).Img.Canvas.Brush.Style:= bsSolid;
  end;
  }

  // From other's nodes, build the 4 vector that connect node i to i+1:
  SetLength(OtherVector, 1+High(Other2D.Nodes));
  for i:= 0 to High(Other2D.Nodes) do begin
    if (i=High(Other2D.Nodes)) then
      nexti:= 0
    else
      nexti:= i+1;
    OtherVector[i].X:= (Other2D.Nodes[nexti].X - Other2D.Nodes[i].X);
    OtherVector[i].Y:= (Other2D.Nodes[nexti].Y - Other2D.Nodes[i].Y);
  end;
  // If ONE of my nodes is inside of the other's polygon, return true and exit:
  for i:= 0 to High(Nodes) do begin
    // Point "A" is inside of a convex poligon "B->C->D->E" if and only if all
    // cross-product B->C ^ B->A (with B->C a vector of the poligon) has the same
    // sign (i.e. point "A" is always at the right -or left- side of the vectors).
    S:= 0;
    ItIsInside:= true;
    for j:= 0 to High(OtherVector) do begin
      CrossProduct:= (Nodes[i].X-Other2D.Nodes[j].X)*OtherVector[j].Y -
                     (Nodes[i].Y-Other2D.Nodes[j].Y)*OtherVector[j].X;
      // Point A is in one side of vector B->C of the polygon, remember the side:
      Side:= Sign(CrossProduct);
      // 0 means point A is aligned with vector B->C, just on the border, so
      // don't count it as a prove it is out or in, just forget about this.
      if (Side<>0) then begin
        if (S=0) then
          // First product non-cero defines the "side" at witch point "A" must
          // always be to be inside the polygon.
          S:= Side
        else if (S <> Side) then begin
          // "A" is not at the right side of this vector, so this point is outside
          // the polygon.

          {
          if (debug=3) then begin
            // Draw empty point:
            TTrack2D(Track).Img.Canvas.Brush.Color:= clWhite;
            TTrack2D(Track).Img.Canvas.Ellipse(Nodes[i].X-2, Nodes[i].Y-2, Nodes[i].X+2, Nodes[i].Y+2);
            Application.ProcessMessages;
          end;
          }

          ItIsInside:= false;
          break;
        end;
      END;
    end;
    // If this node "A" is inside, report the collision and exit.
    if ItIsInside then begin
      {
      if (debug=3) then begin
        // Draw full point:
        TTrack2D(Track).Img.Canvas.Brush.Color:= clRed;
        TTrack2D(Track).Img.Canvas.Ellipse(Nodes[i].X-3, Nodes[i].Y-3, Nodes[i].X+3, Nodes[i].Y+3);
        Application.ProcessMessages;
      end;
      }
      result:= true;
      exit;
    end;
  end;
end;

// Set the inital "position" manually:
procedure TPlayer2D.PlaceHere(x, y, a: double);
begin
  // Set position:
  PosX:=  x;
  PosY:=  y;
  Angle:= a;
  // Stop the movement:
  VelX:= 0;
  VelY:= 0;
  // Use it as the "initial" position:
  SetIniState;
end;

// Move the player from here:
procedure TPlayer2D.HandMove(dx, dy, da: double; repaint: boolean = true);
begin
  // Update my state:
  PosX:=  PosX  + dx;
  PosY:=  PosY  + dy;
  Angle:= Angle + da;
  // Repaint?
  if repaint and Assigned(Track) then begin
    TTrack2D(Track).ClearTrack();
    TTrack2D(Track).Draw();
  end;
end;

// Nodes are the points that form the silouette of the player, as a base, I
// will assume it is a triangle, but each descendant of TPLayer must shape
// its nodes to look ok on screen
procedure TPlayer2D.CalcNodes;
var
  i: integer;
  X, Y: double;
const
  Angles: array[0..2] of double = (0.0, 0.8, -0.8);
begin
  // Nodes is a dynamic array, so I must define size on first run:
  if NodesNeedIni then begin
    SetLength(Nodes, 3);
    NodesNeedIni:= false;
    NodesNeedRecalc:= true;
  end;
  // If nothing changed from last time, don't waste time:
  if NodesNeedRecalc then begin
    for i:= Low(Nodes) to High(Nodes) do begin
      X:= PosX + Size*cos(Angle + Angles[i]*PI);
      Nodes[i].X:= round(X);
      Y:= PosY + Size*sin(Angle + Angles[i]*PI);
      Nodes[i].Y:= round(Y);
    end;
  end;
  // No recalc until a computed param (pos, etc) is changed:
  NodesNeedRecalc:= false;
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TPlayer2D.Draw;
begin
  if not Assigned(Track) then exit;
  inherited;
  CalcNodes;
  // Draw the player's body:
  TTrack2D(Track).Img.Canvas.Pen.Color:= clBlack;
  TTrack2D(Track).Img.Canvas.Brush.Color:= Self.Color;
  TTrack2D(Track).Img.Canvas.Polygon(Nodes);
end;

// Draw the 2D bounding box (debuging):
procedure TPlayer2D.DrawBoxLimits;
begin
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
  TTrack2D(Track).Img.Canvas.Pen.Color:= Self.Color;
  TTrack2D(Track).Img.Canvas.Rectangle(round(MinX), round(MinY), round(MaxX), round(MaxY));
end;
// Draw a big point to remark a position (debugging):
procedure TPlayer2D.DrawBigPoint;
const
  ra: integer = 2;
begin
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsSolid;
  TTrack2D(Track).Img.Canvas.Pen.Color:= clBlack;
  TTrack2D(Track).Img.Canvas.Brush.Color:= Self.Color;
  TTrack2D(Track).Img.Canvas.Ellipse(round(PosX-ra), round(PosY-ra), round(PosX+ra), round(PosY+ra));
end;

// Draw debug information:
procedure TPlayer2D.DrawFuturePoint;
begin
  inherited;
  // Some futures must be hidden at debbug time:
  if DebugThisFuture() then begin
    // This one must be shown:
    TTrack2D(Track).Img.Canvas.Pen.Color:= DebugColor(State);
    TTrack2D(Track).Img.Canvas.MoveTo(round(OldPosX), round(OldPosY));
    TTrack2D(Track).Img.Canvas.LineTo(round(PosX),    round(PosY));
  end;
end;

// Draw circles around a future ending point with radio = scoring.
procedure TPlayer2D.DrawFutureScore(f: TState);
var
  x, y, r: integer;
const
  MaxR: double = 10;
begin
  x:= round(f.ParamByName['PosX'].Value);
  y:= round(f.ParamByName['PosY'].Value);
  // Scale radius so the circles don't grow or shrink too much visually.
  if (FutureScoreScale < 1) then
    FutureScoreScale:= FutureScoreScale * 1.01;
  repeat
    r:= round(1 + Sqrt(f.FinalScore) * FutureScoreScale);
    if (r >= MaxR) then
      FutureScoreScale:= FutureScoreScale * 0.9;
  until (r <= MaxR);
  if r=0 then r:= 1;
  // Draw the circle:
  TTrack2D(Track).Img.Canvas.Pen.Color:= DebugColor(f);
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
  TTrack2D(Track).Img.Canvas.Ellipse(x-r, y-r, x+r, y+r);
end;

// 2D box containing player:
function  TPlayer2D.GetMaxX: double;
begin
  result:= PosX+Size;
end;
function  TPlayer2D.GetMinX: double;
begin
  result:= PosX-Size;
end;
function  TPlayer2D.GetMaxY: double;
begin
  result:= PosY+Size;
end;
function  TPlayer2D.GetMinY: double;
begin
  result:= PosY-Size;
end;


// *******************
// ** SET/GET STUFF **
// *******************

procedure TPlayer2D.SetPosX(v: double);
begin
  State.Param[FPosX].Value:= v;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetPosX: double;
begin
  result:= State.Param[FPosX].Value;
end;

procedure TPlayer2D.SetPosY(v: double);
begin
  State.Param[FPosY].Value:= v;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetPosY: double;
begin
  result:= State.Param[FPosY].Value;
end;

procedure TPlayer2D.SetAngle(v: double);
begin
  State.Param[FAngle].Value:= v;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetAngle: double;
begin
  result:= State.Param[FAngle].Value;
end;

function TPlayer2D.GetOldPosX: double;
begin
  result:= State.Param[FPosX].OldValue;
end;

function  TPlayer2D.GetOldPosY: double;
begin
  result:= State.Param[FPosY].OldValue;
end;

procedure TPlayer2D.SetVelX(v: double);
begin
  State.Param[FVelX].Value:= v;
end;
function  TPlayer2D.GetVelX: double;
begin
  result:= State.Param[FVelX].Value;
end;

procedure TPlayer2D.SetVelY(v: double);
begin
  State.Param[FVelY].Value:= v;
end;
function  TPlayer2D.GetVelY: double;
begin
  result:= State.Param[FVelY].Value;
end;


end.
