unit UTeam;

// TEAM is a group of players that are in a track playing.
// It can have a "Team Intelligence" that will guide the players into the team
// goals

interface

uses Contnrs, SysUtils, Math,
     UPlayer, UGoal;

type
  TTeam = class(TObject)
  private
    // Track I am playing in:
    fTrack: TObject;
    // Player actually "selected":
    fSelectedPlayer: TPlayer;
    // Activate Team Score:
    FUseTeamScore: boolean;
    // The team members are imiagining futures, please call later
    fImaginig: boolean;
    // Witch things motivate this team (affect all its player):
    fGoals: TObjectList;
    fDebugGoals: boolean;
    // Property getters
    procedure SetImagining(v: boolean);
    function  GetPlayer(i: integer): TPlayer;
    function  GetPlayerCount: integer;
    procedure SetSelectedPlayer(v: TPlayer);
    function  GetGoalCount: integer;
  protected
    function  GetGoal(m: integer): TGoal;
  public
    // List of al the players on track (as TObjects):
    fPlayers: TObjectList;
    constructor Create;
    destructor  Destroy; override;
    procedure   IntAddPlayer(v: TPlayer);
    procedure   IntDelPlayer(v: TPlayer; DoFree: boolean = false);
    // Do things on all your team members (players):
    procedure   ResetFutures;
    procedure   ResetState;
    procedure   AddGoal(v: TGoal);
    procedure   DrawGoals; virtual;
    procedure   ScoreGoals(p: integer);
    procedure   TakeFirstStep(sim_time: integer);
    procedure   TakeRandomStep(sim_time: integer);
    procedure   TakeIntelligentStep(sim_time: integer);
    procedure   ThinkAsTeam;
    procedure   Draw;
    property DebugGoals:   boolean read fDebugGoals   write fDebugGoals;
    property Goal[m: integer]: TGoal read GetGoal;
    property GoalCount: integer read GetGoalCount;
    property SelectedPlayer: TPlayer read fSelectedPlayer write SetSelectedPlayer;
    property Player[i: integer]: TPlayer read GetPlayer;
    property PlayerCount: integer read GetPlayerCount;
    property Track: TObject read fTrack write fTrack;
    property Imagining: boolean read fImaginig write SetImagining;
    property UseTeamScore: boolean read FUseTeamScore write FUseTeamScore;
  end;

implementation

uses UTrack, UTrack2D, UListOfFutures, UPlayer2D;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TTeam.Create;
begin
  inherited;
  fPlayers:= TObjectList.create;
  fPlayers.OwnsObjects:= false;
  // Personal goals in life:
  fGoals:= TObjectList.create;
  fGoals.OwnsObjects:= true;
  fSelectedPlayer:= nil;
  fUseTeamScore:= false;
  fImaginig:= false;
  DebugGoals:= false;
end;

destructor TTeam.destroy;
begin
  fPlayers.Free;
  fGoals.Free;
  inherited;
end;


// ********************
// ** THINKING STUFF **
// ********************

// Make the intial random step for this new future:
procedure TTeam.TakeFirstStep(sim_time: integer);
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do begin
    Player[p].TakeFirstDecision;
    Player[p].Simulate(sim_time);
  end;
end;

// Make another future's random step:
procedure TTeam.TakeRandomStep(sim_time: integer);
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do begin
    Player[p].TakeRandomDecision;
    Player[p].Simulate(sim_time);
    ScoreGoals(p);
    // Debugging?
    if Player[p].DebugTraces then
      // Draw a dot so I can follow your imagined future path:
      Player[p].DrawFuturePoint;
  end;
end;

// Make the real -not imagining a future- intelligent step:
procedure TTeam.TakeIntelligentStep(sim_time: integer);
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do begin
    Player[p].TakeIntelligentDecision;
    Player[p].Simulate(sim_time);
    // Force to use the goals in "real life" (so you can eat the sweets)
    ScoreGoals(p);
    // Debugging decision:
    if Player[p].DebugOptions then
      // Draw end points of futures with radius = scorings.
      Player[p].DrawOptions;
  end;
end;

// Use team and player's goals to score a player's future:
procedure TTeam.ScoreGoals(p: integer);
var g: integer;
begin
  // Score personal player's goals first:
  Player[p].ScoreGoals;
  // Add the team's goals on this player:
  for g:= 0 to GoalCount-1 do begin
    Goal[g].ScoreGoal(Player[p]);
    Goal[g].AddScore(Player[p]);
  end;
end;

// CAUTION:
//   Team scoring is not generalized !!!!
//   Only works ok with all players of the same kind and AI_Level 6
//
// Team scoring: Each future is scored by all players individually, but the
// "team" could score the future globally by taking the individual scores and
// "mixing" them in some way.
//
// This new TeamScore is then stored in all player's futures, so future #1 will
// have a diferent MyScore for each player, but the same TeamScore for all.
//
// How a team will score a future is still open, averaging scores from all its
// members is an easy approach, but may be not so optimal.
//
// How a player mix MyScore with TeamScore is open, actually done at TState.
procedure TTeam.ThinkAsTeam;
var
  o, p: integer;
  TeamScore: double;
begin
  if UseTeamScore then begin
    // For each option on the list of futures (all players have same amount of
    // options and they are ordered in the same way IF AI_Level is 6 or +):
    for o:= 0 to Player[0].Futures.OptionCount-1 do begin
      // Average each players' own score for this future:
      // (this is not the only way, there is room for improvements here)
      TeamScore:= 0;
      for p:= 0 to PlayerCount-1 do
        TeamScore:= TeamScore + Player[p].Futures.Option[o].MyScore;
      TeamScore:= TeamScore / PlayerCount;
      // Store this TeamScore on the future[f] of all players:
      for p:= 0 to PlayerCount-1 do
        Player[p].Futures.Option[o].TeamScore:= TeamScore;
    end;
  end;
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TTeam.Draw;
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do
    Player[p].Draw;
end;

// Draw the "Goals" that pushes all team players:
procedure TTeam.DrawGoals;
var
  p, g, mp: integer;
begin
  // Only know how to draw them on a 2D track:
  if (Track is TTrack2D) then begin
    // Draw the individual player's goals:
    for p:= 0 to PlayerCount-1 do
      Player[p].DrawGoals;
    // Draw the team's goals:
    if DebugGoals then begin
      for g:= 0 to GoalCount-1 do begin
        if Goal[g].Drawable then begin
          // I will loop on all players or only one is needed?
          if Goal[g].PersonalizedDrawing then
            // Draw it for each player:
            mp:= PlayerCount-1
          else
            // Draw it just one time:
            mp:= 0;
          // Draw the goal:
          for p:= 0 to mp do
            Goal[g].Draw(Player[p]);
        end;
      end;
    end;
  end;
end;


// **********************
// ** MANAGING PLAYERS **
// **********************

// Accesing a member player:
function TTeam.GetPlayer(i: integer): TPlayer;
begin
  result:= TPlayer(FPlayers[i]);
end;

// Get number of players in the team:
function TTeam.GetPlayerCount: integer;
begin
  result:= FPlayers.Count;
end;

// Add a player to the team:
// Internal, don't call it, instead use Track.AddPlayer
procedure TTeam.IntAddPlayer(v: TPlayer);
begin
  FPlayers.Add(v);
  v.Team:= Self;
  SelectedPlayer:= v;
end;

// Delete a player from the team:
// Internal, don't call it, instead use Track.DelPlayer
procedure TTeam.IntDelPlayer(v: TPlayer; DoFree: boolean = false);
var p: integer;
begin
  p:= FPlayers.IndexOf(v);
  if (p <> -1) then begin
    // Avoid having a dead seelcted player
    if (SelectedPlayer = v) then
      fSelectedPlayer:= nil;
    // Delete player:
    if DoFree then
      Player[p].free;
    FPlayers.Delete(p);
    // Chose another selected pplayer_
    if not Assigned(fSelectedPlayer) and (PlayerCount>0) then
      fSelectedPlayer:= Player[0];
  end;
end;

// Witch player is selected by the user?
procedure TTeam.SetSelectedPlayer(v: TPlayer);
begin
  if fPlayers.IndexOf(v) = -1 then
    raise Exception.create('Selected player must be a team member.');
  fSelectedPlayer:= v;
end;


// *****************
// ** GOALS STUFF **
// *****************

procedure TTeam.AddGoal(v: TGoal);
begin
  fGoals.Add(v);
end;

function TTeam.GetGoalCount: integer;
begin
  result:= fGoals.count;
end;

function TTeam.GetGoal(m: integer): TGoal;
begin
  result:= TGoal(fGoals[m]);
end;


// *************************************
// ** ASK ALL YOUR PLAYERS TO DO THIS **
// *************************************

procedure TTeam.SetImagining(v: boolean);
var i: integer;
begin
  fImaginig:= v;
  for i:= 0 to PlayerCount-1 do
    Player[i].Imagining:= v;
end;

procedure TTeam.ResetFutures;
var i: integer;
begin
  for i:= 0 to PlayerCount-1 do
    Player[i].ResetFutures;
end;

procedure TTeam.ResetState;
var
  p: integer;
begin
  for p:= 0 to PlayerCount-1 do
    Player[p].ResetState;
end;


// *****************
// ** MISCELANEUS **
// *****************






end.
