unit UTrack;

// TRACK is the place where the simulation runs, the chessboard, it knows
// with teams and individual players exists, check for collisions and acts
// as a director on the thinking proccess by syncing all players' future
// creation.

interface

uses ExtCtrls, Contnrs, Forms, StdCtrls,
     UTeam;

type
  TTrack = class(TObject)
  private
    fTeams: TObjectList;
    fSelectedTeam: TTeam;
    // How many milliseconds in the future look ahedad when thinking?
    fFutureMSeconds: integer;
    // How many milliseconds have been the simulation running? And the futures?
    fTimePlayed, fTimeImagined: integer;
    // How many futures will I compute before deciding?
    fMinFutures: integer;
    // Clean players before drawing next frame? (or leave a trail behind)
    FSeeTrails: boolean;
    // Detect collisions?
    FWatchCollisions: boolean;
    // Property getters
    function GetTeam(i: integer): TTeam;
    function GetTeamCount: integer;
    function GetPlayerCount: integer;
  protected
    fPlayers: TObjectList;
    procedure DetectCollisions; virtual; abstract;
    procedure ClearTrack; virtual; abstract;
  public
    constructor Create;
    destructor  Destroy; override;
    // Make all the players to think and move for one tick:
    procedure TakeAStep(CPU_time, sim_time: integer);
    procedure DelTeam(v: TTeam; DoFree: boolean = false);
    // Draw track and all players on theirs positions:
    procedure Draw;
    procedure DrawGoals;
    // Add/Del a player to the selected team:
    procedure AddTeam(v: TTeam);
    procedure AddPlayer(v: TObject);
    procedure DelPlayer(v: TObject);
    property  Team[i: integer]: TTeam read GetTeam;
    property  TeamCount: integer read GetTeamCount;
    property  SelectedTeam: TTeam read fSelectedTeam write fSelectedTeam;
    property  PlayerCount: integer read GetPlayerCount;
    property  SeeTrails: boolean read FSeeTrails write FSeeTrails;
    property  MinFutures: integer read fMinFutures write fMinFutures;
    property  TimePlayed: integer read fTimePlayed;
    property  TimeImagined: integer read fTimeImagined;
    property  FutureMSeconds: integer read FFutureMSeconds write FFutureMSeconds;
    property  WatchCollisions: boolean read FWatchCollisions write FWatchCollisions;
  end;


implementation

uses DateUtils, SysUtils, Graphics, Types, Math,
     UUtils, UPlayer;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TTrack.Create;
begin
  inherited;
  // Teams on track:
  fTeams:= TObjectList.create;
  fTeams.OwnsObjects:= false;
  fSelectedTeam:= nil;
  // Players on the track:
  // Note: It is redundant with team list, but checking collisions needs it.
  fPlayers:= TObjectList.create;
  fPlayers.OwnsObjects:= false;
  // Time counter:
  fTimePlayed:= 0;
  // Futures to compute and for how many seconds:
  fMinFutures:= 100;
  fFutureMSeconds:= 5*1000; // 5s defaul
  // TODO : SeeTrails needs to be deprecated soon!
  SeeTrails:= false;
end;

destructor TTrack.destroy;
begin
  FTeams.Free;
  FPlayers.free;
  inherited;
end;


// ********************
// ** MANAGING TEAMS **
// ********************

function TTrack.GetTeam(i: integer): TTeam;
begin
  result:= TTeam(FTeams[i]);
end;

function TTrack.GetTeamCount: integer;
begin
  result:= FTeams.Count;
end;

// Add a team to the track:
procedure TTrack.AddTeam(v: TTeam);
begin
  FTeams.Add(v);
  v.Track:= Self;
  SelectedTeam:= v;
end;

// Delete a team from the track:
procedure TTrack.DelTeam(v: TTeam; DoFree: boolean = false);
var p: integer;
begin
  p:= FTeams.IndexOf(v);
  if (p <> -1) then begin
    if DoFree then
      Team[p].free;
    FTeams.Delete(p);
  end;
end;


// ******************************
// ** LIST OF PLAYERS ON TRACK **
// ******************************

function TTrack.GetPlayerCount: integer;
begin
  result:= fPlayers.Count;
end;

// Add/Del players using the track instead of the team, as the track needs the
// list of all players on it in order to detect collisions of any 2 players.
procedure TTrack.AddPlayer(v: TObject);
begin
  SelectedTeam.IntAddPlayer(TPlayer(v));
  fPlayers.Add(TPlayer(v));
end;

procedure TTrack.DelPlayer(v: TObject);
begin
  fPlayers.Delete(fPlayers.IndexOf(v));
  SelectedTeam.IntDelPlayer(TPlayer(v), true);
end;


// ********************
// ** THINKING STUFF **
// ********************

// TODO : AutoAdjust MinFutures if a player not having too many options, so
//        before it crash, it is given more time to find a solution.

// Make all players to think, decide, and move (do a complete frame):
// Note: All times in milliseconds.
procedure TTrack.TakeAStep(CPU_time, sim_time: integer);
var
  i, p, t, ticks: integer;
  FoundFutures: integer;
  EndTime: LongWord;
begin
  // Number of ticks needed to complete a future:
  ticks:= Max(2, round(FutureMSeconds/sim_time));
  // Reset time counter into the future:
  fTimeImagined:= 0;
  // Reset the list of futures found:
  for p:= 0 to TeamCount-1 do
    Team[p].ResetFutures;
  // Try to allow 10% of the CPU time to analizing futures and move:
  EndTime:= MillisecondOfTheDay(Now) + round(0.9*CPU_time);
  // Compute futures until time expires and you have the minimun amount:
  FoundFutures:= 0;
  repeat
    // All players enter "imagining mode" (so a new future is created) and take
    // the intial random decision (it must be stored for later use):
    for p:= 0 to TeamCount-1 do begin
      Team[p].Imagining:= true;
      Team[p].TakeFirstStep(sim_time);
    end;
    // Take "sim_time" steps until you get to FutureSeconds
    for t:= 1 to (ticks-1) do begin
      fTimeImagined:= fTimeImagined + sim_time;
      // Ask each player to advance it's future a delta-time:
      for p:= 0 to TeamCount-1 do
        Team[p].TakeRandomStep(sim_time);
      // Detect and react to collisions between any 2 players:
      if WatchCollisions then
        DetectCollisions;
    end;
    // End imagining this future:
    for p:= 0 to TeamCount-1 do
      Team[p].Imagining:= false;
    // Another future computed!
    inc(FoundFutures);
  until (MillisecondOfTheDay(Now) > EndTime) and (FoundFutures > MinFutures);
  //
  // The list of futures is full for all teams and their players, now it is
  // time to (optionally) do a second analysis to get a "TeamScore" for each future.
  for i:= 0 to TeamCount-1 do
    Team[i].ThinkAsTeam;
  // Now players are back to reality.
  // Analize the futures found and decide in a "intelligent" way:
  for p:= 0 to TeamCount-1 do
    Team[p].TakeIntelligentStep(sim_time);
  // Detect "real" collisions between any two players and tell them to react:
  if WatchCollisions then
    DetectCollisions;
  // Screen must be updated if debug info was draw:
  Application.ProcessMessages;
end;

// Draw all players over the track:
procedure TTrack.Draw;
var
  t: integer;
begin
  // TODO: Use triple buffering, one for original track picture, another for
  // adding there trails (dots) and the 3rd to add the players on top.
  // This final 3rd one would also be double-buffered to avoid flickering.
  if not SeeTrails then
    ClearTrack;
  for t:= 0 to TeamCount-1 do
    Team[t].Draw;
end;

procedure TTrack.DrawGoals;
var t: integer;
begin
  for t:= 0 to TeamCount-1 do
    Team[t].DrawGoals;
end;

end.
