unit UTrack2D;

interface

uses ExtCtrls, SysUtils,
     UTrack, UPlayer2D;

type
  TTrack2D = class(TTrack)
  private
    TrackScan: array of array of byte; //0=track, 1=out, 2=red recharge
    // General friction (avoid running too fast):
    FFriction: double;
    // Side friction (avoid sliding):
    FSideFriction: double;
    // Pointer to the track image so i can refresh it:
    FImage: TImage;
    // Image dimensions:
    FImageMaxX, FImageMaxY: integer;
    procedure ScanTrack(Im: TImage);
    function  GetPlayer2D(i: integer): TPlayer2D;
  protected
    procedure DetectCollisions; override;
  public
    CopyImage: TImage;
    constructor Create;
    function    IsInsideImage(i, j: integer): boolean;
    function    IsInsideTrack(i, j: integer): boolean;
    function    Recharge(i, j: integer): boolean;
    procedure   ClearTrack; override;
    property    Player2D[i: integer]: TPlayer2D read GetPlayer2D;
    property    Img: TImage read FImage write ScanTrack;
    property    Friction: double read FFriction write FFriction;
    property    SideFriction: double read FSideFriction write FSideFriction;
  end;

implementation

uses Graphics, Forms, Math,
     UUTils;


// ***************************
// ** ADMNINISTRATIVE STUFF **
// ***************************

constructor TTrack2D.Create;
begin
  inherited;
  FImage:= Nil;
  FFriction:= 0.015;
  FSideFriction:= 0.20;
end;

function TTrack2D.GetPlayer2D(i: integer): TPlayer2D;
begin
  if fPlayers[i] is TPlayer2D then
    result:= TPlayer2D(FPlayers[i])
  else
    raise Exception.create('Track2D contains a player that is not a Player2D.');
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TTrack2D.ScanTrack(Im: TImage);
var
  i, j: integer;
begin
  // Remember it for later use:
  FImage:= Im;
  // Store dimensiones:
  FImageMaxX:= Im.Picture.Width;
  FImageMaxY:= Im.Picture.Height;
  // Scan picture for black pixles:
  SetLength(TrackScan, Im.Picture.Width);
  for i:= Low(TrackScan) to High(TrackScan) do begin
    SetLength(TrackScan[i], Im.Picture.Height);
    for j:= Low(TrackScan[i]) to High(TrackScan[i]) do begin
      TrackScan[i, j]:= 0;
      if (Im.Picture.Bitmap.Canvas.Pixels[i, j] = clBlack) then
        TrackScan[i, j]:= 1; // Outside track
      if (Im.Picture.Bitmap.Canvas.Pixels[i, j] = clRed) then
        TrackScan[i, j]:= 2; // Battery charging zone
    end;
  end;
  // I need a duplicate of the track so I won't paint over the original one.
  CopyImage:= TImage.Create(nil);
  CopyImage.Visible:= false;
  CopyImage.Width:=   Im.Width;
  CopyImage.Height:=  Im.Height;
  CopyImage.Left:=    Im.Left;
  CopyImage.Top:=     Im.Top;
  CopyImage.Picture.Assign(Im.Picture);
end;

procedure TTrack2D.ClearTrack;
begin
  if Assigned(FImage) then
    FImage.Picture.Assign(CopyImage.Picture);
end;


// **********************
// ** SIMULATION STUFF **
// **********************

// Detect collision between 2 players (of any team) and ask them to react:
procedure TTrack2D.DetectCollisions;
var
  i, j, k: integer;
  FromItoJX, FromItoJY, Dist: double;
  DifVelX, DifVelY: double;
  hitIJ, hitJI: boolean;
  Bounce: double;
  PenaltyCoef: double;
begin
  for i:= 0 to PlayerCount-2 do begin
    for j:= (i+1) to PlayerCount-1 do begin
      // First check if a collision is even possible:
      if (Player2D[i].MinX <= Player2D[j].MaxX) and
         (Player2D[i].MaxX >= Player2D[j].MinX) and
         (Player2D[i].MinY <= Player2D[j].MaxY) and
         (Player2D[i].MaxY >= Player2D[j].MinY) then begin
        // A node of Pi is inside Pj's body?
        hitIJ:= Player2D[i].CheckCollision(Player2D[j]);
        if hitIJ then
          hitJI:= false
        else
          hitJI:= Player2D[j].CheckCollision(Player2D[i]);
        if hitIJ or hitJI then begin
          // Ops, players i and j did collide!
          // First thing I need is the vector from Player(i) to Player(j):
          FromItoJX:= Player2D[j].PosX - Player2D[i].PosX;
          FromItoJY:= Player2D[j].PosY - Player2D[i].PosY;
          // Distance from players:
          if (FromItoJX=0) and (FromItoJY=0) then
            Dist:= 1
          else
            Dist:= Sqrt(Sqr(FromItoJX)+Sqr(FromItoJY));
          // Vector distance normalized:
          FromItoJX:= FromItoJX / Dist;
          FromItoJY:= FromItoJY / Dist;
          // Difference of velocities:
          DifVelX:= (Player2D[i].VelX - Player2D[j].VelX);
          DifVelY:= (Player2D[i].VelY - Player2D[j].VelY);
          // Escalar product DiffVel*Dist to sum/dec on each player:
          Bounce:= Max(0, +(DifVelX * FromItoJX) + (DifVelY * FromItoJY));
          // Add bouncing velocity to both players so they now diverge:
          Player2D[i].VelX:= Player2D[i].VelX - FromItoJX * 0.45 * Bounce;
          Player2D[i].VelY:= Player2D[i].VelY - FromItoJY * 0.45 * Bounce;
          Player2D[j].VelX:= Player2D[j].VelX + FromItoJX * 0.45 * Bounce;
          Player2D[j].VelY:= Player2D[j].VelY + FromItoJY * 0.45 * Bounce;
          // Add a penalty:
          PenaltyCoef:= Min(0.8, Max(0, 1-Bounce/20));
          if Player2D[i].Imagining then begin
            // While imagining futures, decrease scoring after a crash to avoid dead:
            Player2D[i].State.ScoreMultiplier:= Player2D[i].State.ScoreMultiplier * PenaltyCoef;
            Player2D[j].State.ScoreMultiplier:= Player2D[j].State.ScoreMultiplier * PenaltyCoef;
            // What if I consider both players "out of track" so their futures
            // end here and lower their scorings dependind on crash velocity?
            // Players will thejn actively avoid crashing. Good!
            if Player2D[i].Imagining then begin
              Player2D[i].Dead:= true;
              Player2D[i].Dead:= true;
              // Test code! Small PenalytyCoef's are equivalent.
              //Player2D[i].State.MyScore:= 0;
              //Player2D[i].State.MyScore:= 0;
            end;
          end else begin
            // While really moving, a crash lowers your health:
            //  not implemented.
          end;
          // Walk back until not touching:
          for k:= 1 to 4 do begin
            Player2D[i].HandMove( -2*FromItoJX, -2*FromItoJY, 0, false);
            Player2D[j].HandMove( +2*FromItoJX, +2*FromItoJY, 0, false);
            if (hitIJ and not Player2D[i].CheckCollision(Player2D[j])) or
               (hitJI and not Player2D[j].CheckCollision(Player2D[i])) then
              break;
          end;
        end;
      end;
    end;
  end;
end;

function TTrack2D.IsInsideImage(i, j: integer): boolean;
begin
  result:= (i>0) and (i<FImageMaxX) and (j>0) and (j<FImageMaxY);
end;

function TTrack2D.IsInsideTrack(i, j: integer): boolean;
begin
  if IsInsideImage(i, j) then
    result:= (TrackScan[i,j]<>1)
  else
    result:= false;
end;

function TTrack2D.Recharge(i, j: integer): boolean;
begin
  if (i>0) and (j>0) and (i<800) and (j<600) then
    result:= (TrackScan[i,j]=2)
  else
    result:= false;
end;


end.
